"""Download necessary wheels and build the `coloraide` wheel."""
import sys
import subprocess
import os
import urllib.request
import urllib.error
import glob
import shutil
import re
import hashlib

# Notebook specific wheels
NOTEBOOK_WHEELS = [
    "https://files.pythonhosted.org/packages/3f/08/83871f3c50fc983b88547c196d11cf8c3340e37c32d2e9d6152abe2c61f7/Markdown-3.7-py3-none-any.whl",  # noqa: E501
    "https://files.pythonhosted.org/packages/09/fb/79a8d27966e90feeeb686395c8b1bff8221727abcbd80d2485841393a955/pymdown_extensions-10.14.1-py3-none-any.whl",  # noqa: E501
]

NOTEBOOK_PYODIDE_PKGS = [
    'pyyaml'
]

# Wheels required in addition to the current project
PLAYGROUND_WHEELS = [
    "https://files.pythonhosted.org/packages/8a/0b/9fcc47d19c48b59121088dd6da2488a49d5f72dacf8262e2790a1d2c7d15/pygments-2.19.1-py3-none-any.whl",  # noqa: E501
    "https://files.pythonhosted.org/packages/1c/9f/7377c2fa9f9bba14f6f4aded34ee30a44d55cba1a7bd9e79d237424e0c13/coloraide-4.3-py3-none-any.whl"  # noqa: E501
]

PLAYGROUND_PYODIDE_PKGS = ['micropip']

MKDOCS_YML = 'mkdocs.yml'

RE_CONFIG = re.compile(r'playground-config.*?\.js')
RE_BUILD = re.compile(r'Successfully built ([-_0-9.a-zA-Z]+?\.whl)')

CONFIG = """\
var colorNotebook = {{
    "playgroundWheels": {},
    "notebookWheels": {},
    "defaultPlayground": "from coloraide_extras.everything import ColorAll as Color\\ncoloraide.__version__\\ncoloraide_extras.__version__\\nColor('color(--ucs 0.27493 0.21264 0.12243 / 1)')"
}}
"""  # noqa: E501

OUTPUT = 'docs/src/markdown/playground/'

NOTEBOOK = {}
for url in NOTEBOOK_WHEELS:
    NOTEBOOK[os.path.join(OUTPUT, url.split('/')[-1])] = url

PLAYGROUND = {}
for url in PLAYGROUND_WHEELS:
    PLAYGROUND[os.path.join(OUTPUT, url.split('/')[-1])] = url


def build_package():
    """Build `coloraide` wheel."""
    cmd = [sys.executable, '-m', 'build', '--wheel', '-o', OUTPUT]

    if sys.platform.startswith('win'):
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            startupinfo=startupinfo,
            shell=False,
            env=os.environ.copy()
        )
    else:
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            shell=False,
            env=os.environ.copy()
        )
    out, _ = process.communicate()
    m = RE_BUILD.search(out.decode('utf-8'))

    return process.returncode, m.group(1) if m else ''


def download_wheel(url, dest):
    """Download a wheel."""

    print('Downloading: {}'.format(url))
    status = 0
    try:
        response = urllib.request.urlopen(url)
        status = response.status
        if status == 200:
            status = 0
            with open(dest, 'wb') as f:
                print('Writing: {}'.format(dest))
                f.write(response.read())
    except urllib.error.HTTPError as e:
        status = e.status

    if status:
        print('Failed to download, recieved status code {}'.format(status))

    return status


if __name__ == "__main__":

    status = 0

    # Clean up all old wheels
    for file in glob.glob(OUTPUT + '*.whl'):
        if file not in NOTEBOOK.keys() and file not in PLAYGROUND.keys():
            os.remove(file)

    for file in glob.glob('docs/theme/playground-config*.js'):
        os.remove(file)

    # Clean up build directory
    if os.path.exists('build'):
        shutil.rmtree('build')

    # Build wheel
    status, package = build_package()
    if not status:
        # Get dependencies
        for file, url in NOTEBOOK.items():
            if os.path.exists(file):
                print('Skipping: {}'.format(file))
                continue
            status = download_wheel(url, file)
            if status:
                break
    if not status:
        for file, url in PLAYGROUND.items():
            if os.path.exists(file):
                print('Skipping: {}'.format(file))
                continue
            status = download_wheel(url, file)
            if status:
                break

    if not status:
        # Build up a list of wheels needed for playgrounds and notebooks
        playground = PLAYGROUND_PYODIDE_PKGS + [os.path.basename(x) for x in PLAYGROUND.keys()] + [package]
        notebook = NOTEBOOK_PYODIDE_PKGS + [os.path.basename(x) for x in NOTEBOOK.keys()] + playground

        # Create the config that specifies which wheels need to be used
        config = CONFIG.format(str(playground), str(notebook)).replace('\r', '').encode('utf-8')
        m = hashlib.sha256()
        m.update(b'playground-config.js')
        m.update(b':')
        m.update(config)
        hsh = m.hexdigest()[0:8]
        with open('docs/theme/playground-config-{}.js'.format(hsh), 'wb') as f:
            f.write(config)

        # Update `mkdocs` source to reference wheel config
        with open(MKDOCS_YML, 'rb') as f:
            mkdocs = f.read().decode('utf-8')
        mkdocs = RE_CONFIG.sub('playground-config-{}.js'.format(hsh), mkdocs)
        with open(MKDOCS_YML, 'wb') as f:
            f.write(mkdocs.encode('utf-8'))

    print("FAILED :(" if status else "SUCCESS :)")
    sys.exit(status)



"""Spectral curve plots."""
import sys
import os
import argparse

sys.path.insert(0, os.getcwd())

from coloraide_extras.interpolate import spectral
from coloraide import algebra as alg
from coloraide.everything import ColorAll as Color
import matplotlib.pyplot as plt

START = 380
STEP = 10
END = 750 + STEP
WL = list(range(START, END, STEP))

def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='plot_reflect', description='Plot reflectance of a given color.')
    parser.add_argument('--color', '-c', action='append', help="Color.")
    parser.add_argument('--title', '-T', default='', help="Title of plot")
    parser.add_argument('--mix', '-m', type=float, default=0.5, help="Mix percentage")
    parser.add_argument('--decomp', '-d', action="store_true", help='Show color decompossed into RGB.')
    parser.add_argument('--resolution', '-r', default="800", help="How densely to render the figure.")
    parser.add_argument('--no-border', '-b', action="store_true", help='Draw no border around the graphed content.')
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()

    colors = args.color

    style = []

    if len(colors) < 1 or len(colors) > 2:
        raise ValueError(f'Need 1 - 2 colors, not {len(colors)}')

    if len(colors) == 1:
        color = Color(args.color[0])
        xyz = color.convert('xyz-d65')[:-1]
        c = alg.matmul(spectral.XYZ_TO_C, xyz)
        r, res1 = spectral.single_constant_xyz_to_reflectance(xyz)

        style = ['solid']
        target = [r]
        plot = [color.convert('srgb').to_string(hex=True)]

        if args.decomp:
            target.append([ri * c[0] for e, ri in enumerate(spectral.REF_R)])
            target.append([ri * c[1] for e, ri in enumerate(spectral.REF_G)])
            target.append([ri * c[2] for e, ri in enumerate(spectral.REF_B)])
            plot.extend(['#ff0000', '#00ff00', '#0000ff'])
            style.extend(['dashed'] * 3)
    else:
        color = Color(args.color[0])
        xyz = color.convert('xyz-d65')[:-1]
        color2 = Color(args.color[1])
        xyz2 = color2.convert('xyz-d65')[:-1]
        r1, res1 = spectral.single_constant_xyz_to_reflectance(xyz)
        r2, res2 = spectral.single_constant_xyz_to_reflectance(xyz2)

        # Apply the Kubelka-Munk mixing
        target = []
        plot = []
        count = 0
        for i in (0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, args.mix):
            t = spectral.nonlinear_luminance_ease(xyz[1], xyz2[1], i)
            size = len(r1)
            r = [0.0] * size
            for i in range(size):
                ks1 = (1 - r1[i]) ** 2 / (2 * r1[i])
                ks2 = (1 - r2[i]) ** 2 / (2 * r2[i])

                # Perform the actual interpolation
                ks = alg.lerp(ks1, ks2, t)

                r[i] = (1 + ks - alg.nth_root(ks ** 2 + 2 * ks, 2))

            if count != 9:
                target.append(r)
                plot.append('#cccccc')
                style.append('dashed')
            else:
                target.append(r)
                xyz1 = spectral.reflectance_to_xyz(r)
                xyz2 = [alg.lerp(r1, r2, t) for r1, r2 in zip(res1, res2)]
                xyz_final = [xyz1[0] + xyz2[0], xyz1[1] + xyz2[1], xyz1[2] + xyz2[2]]
                color3 = Color('xyz-d65', xyz_final)
                plot.append(color3.convert('srgb').to_string(hex=True))
                style.append('solid')

            count += 1

        target.append(r1)
        target.append(r2)
        plot.append(color.convert('srgb').to_string(hex=True))
        plot.append(color2.convert('srgb').to_string(hex=True))
        style.extend(['solid'] * 2)

    # Setup plot for results
    plt.style.use('seaborn-v0_8-darkgrid')

    # Create axes
    ax = plt.axes(
        xlabel='Wavelength',
        ylabel='Reflection'
    )

    # Create titles
    title = args.title
    if not title:
        title = f'Reflectance Curve of {color.to_string()}'
        if not color.in_gamut('srgb'):
            if len(args.color()) == 1:
                plt.suptitle(title)
                ax.set_title('Color out of sRGB gamut, curves may be attenuated')
            else:
                ax.set_title(f'Spectral mix of {color.to_string()} and {color2.to_string()}')
        else:
            ax.set_title(title)
    else:
        ax.set_title(title)

    for e, data in enumerate(zip(plot, target)):
        p, t = data
        plt.plot(
            list(range(START, END, STEP)),
            t,
            color=p,
            marker="",
            linewidth=1.5,
            markersize=2,
            linestyle=style[e],
            antialiased=True
        )

    plt.gcf().set_dpi(200)
    plt.show()

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Test interpolation plugins."""
import unittest
from coloraide_extras.everything import ColorAll as Color
from coloraide import NaN
from . import util


class TestSpectral(util.ColorAsserts, unittest.TestCase):
    """Test spectral color mixing."""

    def test_blue_yellow(self):
        """Test yellow and blue mixing."""

        c1 = Color('#002185')
        c2 = Color('#FCD200')
        expected = [
            Color('color(xyz-d65 0.04777 0.02781 0.22476 / 1)'),
            Color('color(xyz-d65 0.02702 0.03256 0.0952 / 1)'),
            Color('color(xyz-d65 0.03708 0.06349 0.07065 / 1)'),
            Color('color(xyz-d65 0.07141 0.12697 0.07348 / 1)'),
            Color('color(xyz-d65 0.13374 0.22374 0.08054 / 1)'),
            Color('color(xyz-d65 0.22802 0.34531 0.08724 / 1)'),
            Color('color(xyz-d65 0.35287 0.47483 0.09216 / 1)'),
            Color('color(xyz-d65 0.49749 0.59023 0.09487 / 1)'),
            Color('color(xyz-d65 0.6319 0.6679 0.09564 / 1)')
        ]
        for a, b in zip(Color.steps([c1, c2], method='spectral', steps=9), expected):
            self.assertColorEqual(a, b)

    def test_mix_nan(self):
        """Test mixing with NaN."""

        red = Color('red')
        green = Color('green')
        blue = Color('blue')
        green.convert('xyz-d65', in_place=True).set('y', NaN)

        self.assertColorEqual(
            Color.interpolate([red, green, blue], method='spectral')(0.25),
            Color('color(xyz-d65 0.04012 0.03909 0.00367)')
        )

        self.assertColorEqual(
            Color.interpolate([red, green, blue], method='spectral')(0.75),
            Color('color(xyz-d65 0.02697 0.03463 0.0467)')
        )

        red.convert('xyz-d65', in_place=True).set('y', NaN)
        blue.convert('xyz-d65', in_place=True).set('y', NaN)

        self.assertColorEqual(
            Color.interpolate([red, green], method='spectral')(0.5),
            Color('color(xyz-d65 0.13635 -0.05537 0.0189)')
        )

    def test_mix_black(self):
        """Mix with black."""

        self.assertColorEqual(
            Color('black').mix('red', method='spectral'), Color('color(xyz-d65 0.08223 0.04839 0.01747)')
        )
        self.assertColorEqual(
            Color('red').mix('black', method='spectral'), Color('color(xyz-d65 0.08223 0.04839 0.01747)')
        )

    def test_bad_color_space(self):
        """Spectral will only mix in XYZ."""

        with self.assertRaises(ValueError):
            Color('red').mix('blue', method='spectral', space='lab')

    def test_easing(self):
        """Test easing functions."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0)],
                progress=lambda t: 0,
                method='spectral'
            )(1),
            Color('color(xyz-d65 0.41239 0.21264 0.01933 / 1)')
        )

    def test_easing_all(self):
        """Test easing all channels."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0)],
                progress={
                    'all': lambda t: 0,
                },
                method='spectral'
            )(1),
            Color('color(xyz-d65 0.41239 0.21264 0.01933 / 1)')
        )

    def test_easing_channel(self):
        """Test easing specific channels does not work."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0.5)],
                progress={
                    'x': lambda t: 0,
                },
                method='spectral'
            )(1),
            Color('color(xyz-d65 0.18048 0.07219 0.95053 / 0.5)')
        )

    def test_easing_alpha(self):
        """Test easing alpha does work."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0.5)],
                progress={
                    'alpha': lambda t: 0,
                },
                method='spectral'
            )(1),
            Color('color(xyz-d65 0.09024 0.0361 0.47527 / 1)')
        )

    def test_blue_yellow_continuous(self):
        """Test yellow and blue mixing."""

        c1 = Color('#002185')
        c2 = Color('#FCD200')
        expected = [
            Color('color(xyz-d65 0.04777 0.02781 0.22476 / 1)'),
            Color('color(xyz-d65 0.02702 0.03256 0.0952 / 1)'),
            Color('color(xyz-d65 0.03708 0.06349 0.07065 / 1)'),
            Color('color(xyz-d65 0.07141 0.12697 0.07348 / 1)'),
            Color('color(xyz-d65 0.13374 0.22374 0.08054 / 1)'),
            Color('color(xyz-d65 0.22802 0.34531 0.08724 / 1)'),
            Color('color(xyz-d65 0.35287 0.47483 0.09216 / 1)'),
            Color('color(xyz-d65 0.49749 0.59023 0.09487 / 1)'),
            Color('color(xyz-d65 0.6319 0.6679 0.09564 / 1)')
        ]
        for a, b in zip(Color.steps([c1, c2], method='spectral-continuous', steps=9), expected):
            self.assertColorEqual(a, b)

    def test_mix_nan_continuous(self):
        """Test mixing with NaN in continuous mode."""

        red = Color('red')
        green = Color('green')
        blue = Color('blue')
        green.convert('xyz-d65', in_place=True).set('y', NaN)

        self.assertColorEqual(
            Color.interpolate([red, green, blue], method='spectral-continuous')(0.25),
            Color('color(xyz-d65 0.0622 0.04615 0.01112)')
        )

        self.assertColorEqual(
            Color.interpolate([red, green, blue], method='spectral-continuous')(0.75),
            Color('color(xyz-d65 0.02667 0.05355 0.02985)')
        )

    def test_bad_color_space_continuous(self):
        """Spectral continuous will only mix in XYZ."""

        with self.assertRaises(ValueError):
            Color('red').mix('blue', method='spectral-continuous', space='lab')

    def test_easing_continuous(self):
        """Test easing functions."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0)],
                progress=lambda t: 0,
                method='spectral-continuous'
            )(1),
            Color('color(xyz-d65 0.41239 0.21264 0.01933 / 1)')
        )

    def test_easing_all_continuous(self):
        """Test easing all channels."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0)],
                progress={
                    'all': lambda t: 0,
                },
                method='spectral-continuous'
            )(1),
            Color('color(xyz-d65 0.41239 0.21264 0.01933 / 1)')
        )

    def test_easing_channel_continuous(self):
        """Test easing specific channels does not work."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0.5)],
                progress={
                    'x': lambda t: 0,
                },
                method='spectral-continuous'
            )(1),
            Color('color(xyz-d65 0.18048 0.07219 0.95053 / 0.5)')
        )

    def test_easing_alpha_continuous(self):
        """Test easing alpha does work."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0.5)],
                progress={
                    'alpha': lambda t: 0,
                },
                method='spectral-continuous'
            )(1),
            Color('color(xyz-d65 0.09024 0.0361 0.47527 / 1)')
        )




"""Everything but the kitchen sink."""
from __future__ import annotations
from coloraide.everything import ColorAll as Base
from coloraide import stop, hint, NaN
from .spaces.uvw import UVW
from .contrast.contrast_weber import ContrastWeber
from .contrast.contrast_michelson import ContrastMichelson
from .interpolate.spectral import Spectral, SpectralContinuous

__all__ = ("ColorAll", 'NaN', 'stop', 'hint')


class ColorAll(Base):
    """Color class containing all defaults and extra color spaces."""


# If we get out of sync with ColorAide, we may try
# and overwrite a package that used to be here, but
# is now hosted there.
ColorAll.register(
    [
        # Spaces
        UVW(),

        # Delta E

        # Contrast
        ContrastWeber(),
        ContrastMichelson(),

        # Interpolation
        Spectral(),
        SpectralContinuous()

        # Gamut
    ],
    silent=True
)



"""Sanity check that ensures all colors round trip back."""
from coloraide_extras.everything import ColorAll as Base
from coloraide.spaces import HSLish, HSVish
from coloraide import algebra as alg
import pytest
import math


class TestRoundTrip:
    """
    Test round trip conversions for non-achromatic colors.

    A color that does not round trip back likely has a broken conversion.
    It may be possible that some color spaces degrade through conversion.
    Only lower precision for spaces that do not have a high degree of accuracy,
    but they should round trip for their lower precision.
    """

    class Color(Base):
        """Local color object."""

    Color.deregister('space:hpluv')
    Color.deregister('space:ryb-biased')

    SPACES = {k: 6 for k in Color.CS_MAP.keys()}

    COLORS = [
        Color('red'),
        Color('orange'),
        Color('yellow'),
        Color('green'),
        Color('blue'),
        Color('indigo'),
        Color('violet')
    ]

    def assert_round_trip(self, color, space):
        """Cycle through all the other colors and convert to them and back and check the results."""

        c1 = self.Color(color)
        name = c1.to_string(names=True)
        c1.convert(space, in_place=True)
        p1 = self.SPACES[space]
        for space, p2 in self.SPACES.items():
            p = min(p1, p2)
            # Print the color space to easily identify which color space broke.
            c2 = c1.convert(space)
            c2.convert(c1.space(), in_place=True)
            # Catch cases where we are really close to 360 which should wrap to 0
            if c2._space.is_polar():
                if alg.round_half_up(c2.get('hue', nans=False), p) == 360:
                    c2.set('hue', 0)
            # In HSL and HSV spaces particularly, we can get nonsense saturation if lightness
            # is not exactly within 0 - 1 range. Ignore saturation in achromatic cases.
            if isinstance(c2._space, (HSLish, HSVish)) and math.isnan(c2['hue']):
                c2[c2._space.indexes()[1]] = 0.0
            if isinstance(c1._space, (HSLish, HSVish)) and math.isnan(c1['hue']):
                c1[c1._space.indexes()[1]] = 0.0
            # Run rounded string back through parsing in case we hit something like a hue that needs normalization.
            str1 = self.Color(c1.to_string(color=True, fit=False)).to_string(color=True, fit=False, precision=p)
            str2 = self.Color(c2.to_string(color=True, fit=False)).to_string(color=True, fit=False, precision=p)
            # Print failing results for debug purposes
            if str1 != str2:
                print('----- Convert: {} <=> {} -----'.format(c1.space(), space))
                print('Name: ', name)
                print('Original: ', color.to_string(color=True, fit=False, precision=p))
                print(c1.space() + ': ', str1, c1[:])
                print(space + ': ', str2, c2[:])
                assert str1 == str2

    @pytest.mark.parametrize('space', SPACES)
    def test_round_trip(self, space):
        """Test round trip."""

        for c in self.COLORS:
            self.assert_round_trip(c, space)


class TestAchromaticRoundTrip(TestRoundTrip):
    """
    Test achromatic round trip.

    For convenience, we determine achromatic colors and set hues to `NaN`.
    This can cause a slight drop in precision.
    """

    class Color(Base):
        """Local color object."""

    Color.deregister('space:hpluv')
    Color.deregister('space:ryb')
    Color.deregister('space:ryb-biased')

    SPACES = {k: 5 for k in Color.CS_MAP.keys()}
    # Precision just isn't as high for these in achromatic region
    # but it is good enough for practical purposes.

    COLORS = [
        Color('darkgrey'),
        Color('lightgrey'),
        Color('gray'),
        Color('black'),
        Color('white')
    ]

    @pytest.mark.parametrize('space', SPACES)
    def test_round_trip(self, space):
        """Test round trip."""

        for c in self.COLORS:
            self.assert_round_trip(c, space)



"""
Approach based off of Spectral.js library.

https://github.com/rvanwijnen/spectral.js
---
MIT License

 Copyright (c) 2023 Ronald van Wijnen

 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
---
Modified by Isaac Muse
- Recalculated D65 illuminated CMFs and R, G, and B curves with higher precision
  and generated them with our exact RGB matrix and CMF calculations.
- Eliminated need for C, M, Y curves as R, G, B covered the full gamut and is
  how the eye works.
- Use least squares matrix to calculate the concentrations.
- Added a few normalizations for when curve exceeds with higher gamuts.
- Calculate residual and add back when calculating colors beyond the spectral gamut.
"""
from __future__ import annotations
import math
from coloraide import algebra as alg
from coloraide.types import Vector
from coloraide.interpolate import Interpolator, Interpolate
from coloraide.interpolate.linear import InterpolatorLinear
from coloraide.interpolate.continuous import InterpolatorContinuous
from typing import Any, Mapping, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from coloraide.color import Color

EPSILON = 1e-12

SPACE = 'xyz-d65'

X_BAR = [
    6.4691998957636e-05, 0.00021940989981324543, 0.0011205743509342526, 0.003766613411711093,
    0.011880553603799004, 0.023286442419177128, 0.034559418196974744, 0.03722379011620067,
    0.03241837610914861, 0.021233205609381033, 0.01049099076854208, 0.003295837579793109,
    0.0005070351633801336, 0.0009486742057141478, 0.006273718099831793, 0.0168646241897775,
    0.028689649025980982, 0.04267481246917313, 0.05625474813113776, 0.06947039726771581,
    0.08305315169982916, 0.08612609630022569, 0.09046613768477696, 0.08500386505912774,
    0.0709066691074488, 0.050628891637364504, 0.03547396188526398, 0.021468210259706556,
    0.012516456761911689, 0.006804581639016529, 0.0034645657946526308, 0.0014976097506959416,
    0.000769700480928044, 0.00040736805813154517, 0.0001690104031613905, 9.5224515036545e-05,
    4.903098729584765e-05, 1.999614922216866e-05
]

Y_BAR = [
    1.8442894439676924e-06, 6.205323586516486e-06, 3.100960467994158e-05, 0.00010474838492692305,
    0.00035364052995383256, 0.0009514714056444336, 0.0022822631748317997, 0.004207329043473007,
    0.006688798371901364, 0.009888396019356503, 0.015249451449631114, 0.02141831094497228,
    0.033422930157506775, 0.05131001349185122, 0.070402083939949, 0.0878387072603517,
    0.0942490536184086, 0.09795667027189314, 0.09415218568626084, 0.08678102374867531,
    0.07885653386320132, 0.06352670262035551, 0.05374141675682006, 0.042646064357411986,
    0.03161734927927079, 0.020885205921391023, 0.01386011013601517, 0.008102640203839865,
    0.004630102258802989, 0.0024913800051319097, 0.0012593033677377537, 0.0005416465221680035,
    0.00027795289200670086, 0.00014710806738544828, 6.103274729272546e-05, 3.43873229523396e-05,
    1.7705986005253943e-05, 7.220974912993785e-06
]

Z_BAR = [
    0.00030501714763797594, 0.0010368066663574284, 0.0053131363323992, 0.01795439258995359,
    0.05707758153454857, 0.11365161893628682, 0.17335872618354975, 0.19620657555865664,
    0.18608237070629596, 0.13995047538320737, 0.08917452942686492, 0.04789621135170755,
    0.02814562539579518, 0.01613766229505142, 0.0077591019215213644, 0.00429614837366175,
    0.002005509212215613, 0.0008614711098801786, 0.0003690387177652434, 0.000191428728857372,
    0.00014955558589748968, 9.231092851042412e-05, 6.813491823368628e-05, 2.8826365569622417e-05,
    1.5767182055279397e-05, 3.940604102707517e-06, 1.5840125869731628e-06, 0.0,
    0.0, 0.0, 0.0, 0.0,
    0.0, 0.0, 0.0, 0.0,
    0.0, 0.0
]

REF_R = [
    0.031560573775605893, 0.03155207183104386, 0.03151482154968238, 0.031331804497015225,
    0.03067298577184263, 0.028648047698799695, 0.02464504070475776, 0.019296075366599053,
    0.014206661222182781, 0.010294260887878548, 0.007619146052136205, 0.005898041083471917,
    0.004823324778097049, 0.004229874834993319, 0.004059917129869861, 0.004353369559408238,
    0.005343442596964287, 0.007691720100996724, 0.013596979573625545, 0.03169754426617544,
    0.10786119635570235, 0.46381260316918005, 0.8470554052715022, 0.9431854093936334,
    0.9688621506964221, 0.978030667473553, 0.9820436438543185, 0.9839236237187765,
    0.9848454841545076, 0.9852942758147769, 0.9855072952200589, 0.9856050715401196,
    0.985653849933904, 0.9856776850342464, 0.9856883918065151, 0.9856936646904471,
    0.9856958798486362, 0.9856965214642008
]

REF_G = [
    0.009556074754410449, 0.00955815801112031, 0.009567324543584887, 0.009612912628991555,
    0.009783709039599076, 0.010378622705444651, 0.012002645237574217, 0.01609777214724728,
    0.02670619022321924, 0.0595555440190878, 0.18603982653435341, 0.5705798201163349,
    0.8614677683992376, 0.945879089766979, 0.9704654864738727, 0.9784136302841248,
    0.9795890314109317, 0.9755335369083287, 0.9622887553974709, 0.9231215745128396,
    0.7934340189439301, 0.45927013590622906, 0.18557410366813132, 0.08817749599543823,
    0.05436302287555966, 0.04062884470419337, 0.03422152042906518, 0.03111857909223853,
    0.029570889829271474, 0.028810873929642933, 0.02844862712632218, 0.028282030165508287,
    0.028198837641326635, 0.028158165525883294, 0.028139891012812057, 0.028130890157372412,
    0.02812710867111673, 0.028126013351616186
]

REF_B = [
    0.9794047525027293, 0.9794007068438306, 0.9793829034709327, 0.9792943649462249,
    0.9789630146091524, 0.9778144666945839, 0.9747243211343526, 0.9671984823444855,
    0.9490796575310576, 0.9008501289411428, 0.7631504454607034, 0.4659221716451323,
    0.20126328044854624, 0.08775244134130433, 0.045717679329214, 0.0284706050524961,
    0.020527176757419385, 0.01653027923153466, 0.014513510721861633, 0.013600350864389432,
    0.013360425877586513, 0.013548894315131066, 0.013959435637084183, 0.014443425575408897,
    0.014885444061693232, 0.015225429698892456, 0.015459284816230268, 0.015601802646075802,
    0.015682487124966893, 0.01572487643217746, 0.01574581087393412, 0.015755612330020707,
    0.015760544391039022, 0.01576296374570002, 0.015764052556781927, 0.01576458922659668,
    0.015764814770760138, 0.015764880108381563
]

SIZE = len(X_BAR)

REFLECTANCE = [REF_R, REF_G, REF_B]

C_TO_XYZ = [
    [0.4123907992659593, 0.3575843393838777, 0.1804807884018343],
    [0.21263900587151033, 0.7151686787677553, 0.07219231536073373],
    [0.019330818715591832, 0.11919477979462595, 0.9505321522496605]
]
# Below is the Moore Penrose pseudo inverse used to calculate the least squares
# to get the concentrations for an XYZ value.
XYZ_TO_C = [
    [3.2409699419045253, -1.537383177570097, -0.4986107602930039],
    [-0.9692436362808824, 1.8759675015077237, 0.041555057407175744],
    [0.055630079696993795, -0.20397695888897688, 1.0569715142428786]
]


def nonlinear_luminance_ease(l1: float, l2: float, t: float) -> float:
    """
    Scale the transition in a non-linear manner.

    As our curves are calculated directly from light data, there is no
    context of true pigment characteristics. Particularly, our results
    will mix linearly while this is not true for paint.

    Based on the luminance difference in the two colors, the proportion
    of mixing (the dominance of one color over another) will be altered.

    Essentially, a color with more luminance will dominate the mix.
    """

    if l1 <= 0.0:
        l1 = EPSILON
    if l2 <= 0.0:
        l2 = EPSILON

    t1 = l1 * (1 - t) ** 2
    t2 = l2 * t ** 2
    return t2 / (t1 + t2)


def single_constant_xyz_to_reflectance(xyz: Vector) -> tuple[Vector, Vector]:
    """
    Linear sRGB to a reflectance.

    Use least squares method to calculate concentration of our primary colors with the XYZ color.
    If within the sRGB gamut, we will not get non-negative results (after clipping some floating point noise).

    For out of gamut colors, if our concentration concentration are clipped to 0 - 1. Lastly, we ensure the
    final reflectance result is never zero with a small epsilon.

    Because out of gamut colors may be attenuated due to constraints on concentrations, we must
    also calculate the residual, left over weights of our XYZ value and return them as well. We can
    use the residual later to better approximate colors out of gamut by adding them back in.
    """

    c = alg.matmul_x3(XYZ_TO_C, xyz, dims=alg.D2_D1)
    c = [alg.clamp(i, 0.0, 1.0) for i in c]
    r = [alg.clamp(sum([c[j] * p[i] for j, p in enumerate(REFLECTANCE)]), EPSILON) for i in range(SIZE)]
    xyz2 = reflectance_to_xyz(r)
    return r, [xyz[0] - xyz2[0], xyz[1] - xyz2[1], xyz[2] - xyz2[2]]


def reflectance_to_xyz(r: Vector) -> Vector:
    """Convert the reflectance value to an XYZ value."""

    return [alg.vdot(r, X_BAR), alg.vdot(r, Y_BAR), alg.vdot(r, Z_BAR)]


def spectral_mix(xyz1: Vector, xyz2: Vector, t: float) -> Vector:
    """Interpolate two colors applying Kubelka-Munk theory."""

    # Convert the colors into a reflectance curve
    r1, res1 = single_constant_xyz_to_reflectance(xyz1)
    r2, res2 = single_constant_xyz_to_reflectance(xyz2)

    # Adjust the mixing based on when color dominates due to luminance.
    t = nonlinear_luminance_ease(xyz1[1], xyz2[1], t)

    # Apply the Kubelka-Munk mixing
    r = [0.0] * SIZE
    for i in range(SIZE):
        # Reflectance to k/s
        ks1 = (1 - r1[i]) ** 2 / (2 * r1[i])
        ks2 = (1 - r2[i]) ** 2 / (2 * r2[i])

        # Perform the actual interpolation
        ks = alg.lerp(ks1, ks2, t)

        # Back to reflectance
        r[i] = 1 + ks - alg.nth_root(ks ** 2 + 2 * ks, 2)

    # Convert the reflection back to XYZ and add back in any residual
    xyz1 = reflectance_to_xyz(r)
    xyz2 = [alg.lerp(r1, r2, t) for r1, r2 in zip(res1, res2)]
    return [xyz1[0] + xyz2[0], xyz1[1] + xyz2[1], xyz1[2] + xyz2[2]]


class InterpolatorSpectralContinuous(InterpolatorContinuous):
    """Interpolate with continuous piecewise."""

    def mix(self, a: Vector, b: Vector, t: float) -> Vector:
        """Perform a Kubelka-Munk mix."""

        return spectral_mix(a, b, t)

    def interpolate(
        self,
        point: float,
        index: int
    ) -> Vector:
        """Interpolate."""

        # Interpolate between the values of the two colors for each channel.
        channels = []
        idx = index - 2 if index == self.length else index - 1

        # Handle spectral interpolation
        c1, c2 = self.coordinates[idx:idx + 2]
        channels = self.mix(
            [0.0 if math.isnan(i) else i for i in c1[:-1]],
            [0.0 if math.isnan(i) else i for i in c2[:-1]],
            self.ease(point, 0)
        )
        channels.append(alg.lerp(c1[-1], c2[-1], self.ease(point, len(c1) - 1)))
        return channels

    def ease(self, t: float, channel_index: int) -> float:
        """Provide a progression time and channel index."""

        progress = None
        if self.current_easing is not None:
            # Do we have an easing function, or mapping with a channel easing function?
            name = self.channel_names[channel_index]
            if isinstance(self.current_easing, Mapping):
                progress = self.current_easing.get(name) if name == 'alpha' else None
                if progress is None:
                    progress = self.current_easing.get('all')
            else:
                progress = self.current_easing

        return progress(t) if progress is not None else t


class InterpolatorSpectralLinear(InterpolatorLinear):
    """Interpolate multiple ranges of colors using linear, Piecewise interpolation."""

    def mix(self, a: Vector, b: Vector, t: float) -> Vector:
        """Perform a Kubelka-Munk mix."""

        return spectral_mix(a, b, t)

    def interpolate(
            self,
            point: float,
            index: int
        ) -> Vector:
            """Interpolate."""

            i = (index - 1) * 2

            # Apply spectral interpolation
            c1, c2 = self.coordinates[i:i + 2]
            aidx = len(c1) - 1
            for i in range(len(c1)):
                a, b = c1[i], c2[i]
                if math.isnan(a) and math.isnan(b):
                    if i != aidx:
                        c1[i], c2[i] = 0.0, 0.0
                elif math.isnan(a):
                    c1[i] = b
                elif math.isnan(b):
                    c2[i] = a
            return (
                self.mix(c1[:-1], c2[:-1], self.ease(point, 0)) +
                [alg.lerp(c1[aidx], c2[aidx], self.ease(point, aidx))]
            )

    def ease(self, t: float, channel_index: int) -> float:
        """Provide a progression time and channel index."""

        progress = None
        if self.current_easing is not None:
            # Do we have an easing function, or mapping with a channel easing function?
            name = self.channel_names[channel_index]
            if isinstance(self.current_easing, Mapping):
                progress = self.current_easing.get(name) if name == 'alpha' else None
                if progress is None:
                    progress = self.current_easing.get('all')
            else:
                progress = self.current_easing

        return progress(t) if progress is not None else t


class Spectral(Interpolate):
    """Linear interpolation plugin."""

    NAME = "spectral"

    def interpolator(self, *args: Any, **kwargs: Any) -> Interpolator:
        """Return the linear interpolator."""

        return InterpolatorSpectralLinear(*args, **kwargs)

    def get_space(self, space: str | None, color_cls: type[Color]) -> str:
        """Filter specified spaces."""

        if space is None or space == SPACE:
            return SPACE

        raise ValueError(f"Only '{SPACE}' is supported for '{self.NAME}' interpolation")


class SpectralContinuous(Interpolate):
    """Linear interpolation plugin."""

    NAME = "spectral-continuous"

    def interpolator(self, *args: Any, **kwargs: Any) -> Interpolator:
        """Return the linear interpolator."""

        return InterpolatorSpectralContinuous(*args, **kwargs)

    def get_space(self, space: str | None, color_cls: type[Color]) -> str:
        """Filter specified spaces."""

        if space is None or space == SPACE:
            return SPACE

        raise ValueError(f"Only '{SPACE}' is supported for '{self.NAME}' interpolation")



"""
Weber color contrast.

https://en.wikipedia.org/wiki/Contrast_(vision)#Weber_contrast.
"""
from __future__ import annotations
from coloraide.contrast import ColorContrast
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from coloraide import Color

MAX = 50000


class ContrastWeber(ColorContrast):
    """Weber contrast."""

    NAME = "weber"

    @classmethod
    def contrast(cls, color1: Color, color2: Color, **kwargs: Any) -> float:
        """Contrast."""

        lum1 = max(color1.luminance(), 0.0)
        lum2 = max(color2.luminance(), 0.0)

        if lum1 > lum2:
            lum2, lum1 = lum1, lum2

        if lum1 == 0:
            return MAX

        return (lum2 - lum1) / lum1



"""
Michelson color contrast.

https://en.wikipedia.org/wiki/Contrast_(vision)#Michelson_contrast.
"""
from __future__ import annotations
from coloraide.contrast import ColorContrast
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from coloraide import Color


class ContrastMichelson(ColorContrast):
    """Michelson contrast."""

    NAME = "michelson"

    @classmethod
    def contrast(cls, color1: Color, color2: Color, **kwargs: Any) -> float:
        """Contrast."""

        lum1 = max(color1.luminance(), 0.0)
        lum2 = max(color2.luminance(), 0.0)

        if lum1 > lum2:
            lum2, lum1 = lum1, lum2

        d = lum2 + lum1

        if d == 0:
            return 0

        return (lum2 - lum1) / d



"""
Calculate reflection curves.

Approaches comes from Scott Burns.
http://scottburns.us/matlab-octave-and-python-source-code-for-refl-recon-chrom-adapt/
"""
import sys
import os
import math
import matplotlib.pyplot as plt

sys.path.insert(0, os.getcwd())

from coloraide import algebra as alg  # noqa: E402
from coloraide.everything import ColorAll as Color  # noqa: E402
from coloraide.cmfs import CIE_1931_2DEG as CMF  # noqa: E402

targets = ('#ff0000', '#00ff00', '#0000ff')
space = 'srgb'
START = 380
STEP = 10
END = 750 + STEP
MAX_ITERS = 50
EPSILON = 1e-15
METHOD = 3

D65 = {
    300: 0.034100,
    305: 1.664300,
    310: 3.294500,
    315: 11.765200,
    320: 20.236000,
    325: 28.644700,
    330: 37.053500,
    335: 38.501100,
    340: 39.948800,
    345: 42.430200,
    350: 44.911700,
    355: 45.775000,
    360: 46.638300,
    365: 49.363700,
    370: 52.089100,
    375: 51.032300,
    380: 49.975500,
    385: 52.311800,
    390: 54.648200,
    395: 68.701500,
    400: 82.754900,
    405: 87.120400,
    410: 91.486000,
    415: 92.458900,
    420: 93.431800,
    425: 90.057000,
    430: 86.682300,
    435: 95.773600,
    440: 104.865000,
    445: 110.936000,
    450: 117.008000,
    455: 117.410000,
    460: 117.812000,
    465: 116.336000,
    470: 114.861000,
    475: 115.392000,
    480: 115.923000,
    485: 112.367000,
    490: 108.811000,
    495: 109.082000,
    500: 109.354000,
    505: 108.578000,
    510: 107.802000,
    515: 106.296000,
    520: 104.790000,
    525: 106.239000,
    530: 107.689000,
    535: 106.047000,
    540: 104.405000,
    545: 104.225000,
    550: 104.046000,
    555: 102.023000,
    560: 100.000000,
    565: 98.167100,
    570: 96.334200,
    575: 96.061100,
    580: 95.788000,
    585: 92.236800,
    590: 88.685600,
    595: 89.345900,
    600: 90.006200,
    605: 89.802600,
    610: 89.599100,
    615: 88.648900,
    620: 87.698700,
    625: 85.493600,
    630: 83.288600,
    635: 83.493900,
    640: 83.699200,
    645: 81.863000,
    650: 80.026800,
    655: 80.120700,
    660: 80.214600,
    665: 81.246200,
    670: 82.277800,
    675: 80.281000,
    680: 78.284200,
    685: 74.002700,
    690: 69.721300,
    695: 70.665200,
    700: 71.609100,
    705: 72.979000,
    710: 74.349000,
    715: 67.976500,
    720: 61.604000,
    725: 65.744800,
    730: 69.885600,
    735: 72.486300,
    740: 75.087000,
    745: 69.339800,
    750: 63.592700,
    755: 55.005400,
    760: 46.418200,
    765: 56.611800,
    770: 66.805400,
    775: 65.094100,
    780: 63.382800
}

if space == 'srgb':
    from coloraide.spaces.srgb_linear import XYZ_TO_RGB
elif space == 'display-p3':
    from coloraide.spaces.display_p3_linear import XYZ_TO_RGB
elif space == 'a98-rgb':
    from coloraide.spaces.a98_rgb_linear import XYZ_TO_RGB
elif space == 'rec2020':
    XYZ_TO_RGB = [
        [1.7249510969855462, -0.36225403975096726, -0.2545649705534804],
        [-0.6692669194491035, 1.6225096012804534, 0.012487041279557484],
        [0.018196517838035248, -0.04443708810957042, 0.9431475044937067]
    ]
else:
    raise ValueError('Cannot perfrom operation on this space')


A = [[], [], []]
for r in range(START, END, STEP):
    A[0].append(CMF[r][0])
    A[1].append(CMF[r][1])
    A[2].append(CMF[r][2])

W = [D65[r] for r in range(START, END, STEP)]

value = alg.matmul(A, alg.diag(W))
w = alg.vdot(A[1], W)

# Illuminant referenced CMFs
T = alg.divide(value, w)

# Make in relation to the RGB color space
TRGB = alg.matmul(XYZ_TO_RGB, T)

# Array of finite differencing constants
size = len(W)
fill = {-1: -2, 0: 4, 1: -2}
D = [[fill.get(i - j, 0) for j in range(size)] for i in range(size)]
D[0][0] = 2
D[size - 1][size - 1] = 2

# Vectorize some functions we will use
exp = alg.vectorize2(math.exp, params=1)
tanh = alg.vectorize2(math.tanh, params=1)


def _exp(x):
    """This can overflow, so clamp the result to zero."""
    try:
        return math.exp(x)
    except OverflowError:
        return 0


def _sech2(x):
    """This can overflow, so clamp the result to zero."""
    try:
        return 1 / (math.cosh(x) ** 2)
    except OverflowError:
        return 0

sech2 = alg.vectorize2(_sech2, params=1)
exp = alg.vectorize2(_exp, params=1)


def method_2(d, cmfs_w, XYZ_w, max_iterations=MAX_ITERS, tolerance=EPSILON):
    """
    Estimate reflectance curve.

    Computes a reconstructed spectral reflectance curve from
    an illuminant-referenced triplet of tristimulus values
    representing a real color (within the spectral locus).

    Parameters
    ----------
    `d`: nxn array of finite differencing constants, obtained
           from function `method_2_prep.py`. (n is the number
           of wavelength bins (rows) used in the CMFs.)
    `cmfs_w`: nx3 array of illuminant-w-referenced CMFs,
           obtained from function `method_2_prep.py`.
    `XYZ_w`: a three element vector of illum-w-referenced
           tristimulus values in the 0 <= Y <= 1 convention range.
    max_iterations: upper limit on iterations (defaults to 20).
    tolerance: determines when the linear equations for a
           stationary point are satisfied (default is 1e-8).

    Returns
    -------
    nx1 vector of spectral reflectance values (or zeros if failure).

    Notes
    -----
    The input parameters d and `cmfs_w` can be generated by function
    `method_2_prep.py`. For a given illuminant, `method_2_prep`
    needs to be called only once. Multiple calls to `method_2`
    may then follow for various tristimulus values. This is a
    typical use case, and avoids repeated computation of d and
    `cmfs_w`.

    The XYZ_w supplied to `method_2` must fall strictly within the
    spectral locus (i.e., be a real color), or the method will fail
    and return a vector of zeros.

    The reflectance curve generated will always have positive
    values, but may have some values >1 (which can be interpreted
    as an emissive source divided by the illuminant). To obtain
    reflectance curves strictly in the range 0 to 1, representing
    valid object reflectances, use `method_3.py` instead.

    `method_2` is based on "method 2" of Reference [1]. It also
    corresponds to method "LLSS" (Least Log Slope Squared) of
    Reference [2].

    References
    ----------
    [1] Burns SA. Numerical methods for smoothest reflectance
        reconstruction. Color Research & Application, Vol 45,
        No 1, 2020, pp 8-21.
    [2] Generating Reflectance Curves from sRGB Triplets, 2015
        http://scottburns.us/reflectance-curves-from-srgb/#LLSS

    Acknowledgements
    ----------------
    Thanks to Adam J. Burns and Mark (`kram1032`) for assistance
    with translating from MATLAB to Python.
    """

    n = len(cmfs_w)
    z = alg.zeros(n)
    lamb = alg.zeros(3)
    for _ in range(max_iterations):
        r = exp(z)
        dr = alg.diag(r)
        dra = alg.matmul(dr, cmfs_w)
        v = alg.matmul(dra, lamb)
        p1 = alg.add(alg.matmul(d, z), v)
        p2 = alg.subtract(alg.matmul(alg.transpose(cmfs_w), r), XYZ_w)
        F = p1 + p2
        p1 = alg.add(d, alg.diag(v))
        p2 = dra
        p3 = alg.transpose(dra)
        p4 = alg.zeros((3, 3))
        for x1, y1 in zip(p1, p2):
            x1.extend(y1)
        for x1, y1 in zip(p3, p4):
            x1.extend(y1)
        p1.extend(p3)
        J = p1
        try:
            try:
                delta = alg.solve(J, alg.multiply(F, -1))
            except Exception:
                pi = alg.pinv(J)
                delta = alg.dot(pi, alg.multiply(F, -1))
        except Exception:
            print('Ill-conditioned or singular linear system detected.')
            print(f'Check to make sure XYZ_w {XYZ_w} is within the spectral locus.')
            return alg.zeros(n)
        z = alg.add(z, delta[:n])
        lamb = alg.add(lamb, delta[n:])
        if alg.all([abs(f) < tolerance for f in F]):
            return exp(z)
    print('Maximum number of iterations reached.')
    print('Check to make sure XYZ_w is within the spectral locus.')
    return alg.zeros(n)


def method_3(d, cmfs_w, XYZ_w, max_iterations=MAX_ITERS, tolerance=EPSILON):
    """
    Estimate reflectance curve.

    Computes a reconstructed spectral reflectance curve from
    an illuminant-referenced triplet of tristimulus values
    representing an object color (within the object color solid).

    Parameters
    ----------
    `d`: nxn array of finite differencing constants, obtained
           from function `method_3_prep.py`. (n is the number
           of wavelength bins (rows) used in the CMFs.)
    `cmfs_w`: nx3 array of illuminant-w-referenced CMFs,
           obtained from function `method_3_prep.py`.
    `XYZ_w`: a three element vector of illum-w-referenced
           tristimulus values in the 0 <= Y <= 1 convention range.
    max_iterations: upper limit on iterations (defaults to 20).
    tolerance: determines when the linear equations for a
           stationary point are satisfied (default is 1e-8).

    Returns
    -------
    nx1 vector of spectral reflectance values (or zeros if failure).

    Notes
    -----
    The input parameters d and `cmfs_w` can be generated by function
    `method_3_prep.py`. For a given illuminant, `method_3_prep`
    needs to be called only once. Multiple calls to `method_3`
    may then follow for various tristimulus values. This is a
    typical use case, and avoids repeated computation of d and
    `cmfs_w`.

    The XYZ_w supplied to `method_3` must fall strictly within the
    object color solid (i.e., be a valid object color that can be
    represented by a reflectance curve strictly between 0 and 1),
    or the method will fail and return a vector of zeros.

    The reflectance curve generated will always have values in
    the range 0 to 1, corresponding to valid object colors. If
    it is desired to expand the applicability of the method to
    any real color within the spectral locus (including emissive
    colors), use `method_2.py` instead.

    `method_3` is based on "method 3" of Reference [1]. It also
    corresponds to method "LHTSS" (Least Hyperbolic Tangent Slope
    Squared) of Reference [2].

    References
    ----------
    [1] Burns SA. Numerical methods for smoothest reflectance
        reconstruction. Color Research & Application, Vol 45,
        No 1, 2020, pp 8-21.
    [2] Generating Reflectance Curves from sRGB Triplets, 2015
        http://scottburns.us/reflectance-curves-from-srgb/#LHTSS

    Acknowledgements
    ----------------
    Thanks to Adam J. Burns and Mark (`kram1032`) for assistance
    with translating from MATLAB to Python.

    """

    n = len(cmfs_w)  # number of rows
    z = alg.zeros(n)
    lam = alg.zeros(3)
    for _ in range(max_iterations):
        r = alg.divide(alg.add(tanh(z), 1), 2)
        s2 = sech2(z)
        d1 = alg.diag(alg.divide(s2, 2))
        d1a = alg.dot(d1, cmfs_w)
        d2 = alg.diag(alg.multiply(s2, tanh(z)))

        p1 = alg.add(alg.dot(d, z), alg.dot(d1a, lam))
        p2 = alg.subtract(alg.dot(alg.transpose(cmfs_w), r), XYZ_w)
        f = p1 + p2

        p1 = alg.subtract(d, alg.diag(alg.dot(alg.dot(d2, cmfs_w), lam)))
        p2 = d1a
        p3 = alg.transpose(d1a)
        p4 = alg.zeros((3, 3))
        for x1, y1 in zip(p1, p2):
            x1.extend(y1)
        for x1, y1 in zip(p3, p4):
            x1.extend(y1)
        p1.extend(p3)
        j = p1

        try:
            try:
                delta = alg.solve(j, alg.multiply(-1, f))
            except Exception:
                pi = alg.pinv(j)
                delta = alg.dot(pi, alg.multiply(f, -1))
        except Exception:
            print('Ill-conditioned or singular linear system detected.')
            print(f'Check to make sure XYZ_w {XYZ_w} is within the spectral locus.')
            return alg.zeros(n)
        z = alg.add(z, delta[0:n])
        lam = alg.add(lam, delta[n:n+3])
        if alg.all([abs(_f) < tolerance for _f in f]):
            return alg.divide(alg.add(tanh(z), 1), 2)
    print('Maximum number of iterations reached.')
    print('Check to make sure XYZ_w is within the object color solid.')
    return alg.zeros(n)

pcolors = []
rcolors = []
# Estimate curves for the specified colors red, green, and blue
rho = []
for c in ([1, 0, 0], [0, 1, 0], [0, 0, 1]):
    if METHOD == 3:
        r = method_3(D, alg.transpose(TRGB), c)
    elif METHOD == 2:
        r = method_2(D, alg.transpose(TRGB), c)
    else:
        raise ValueError('Unrecognized method')
    rho.append(r)

# Calculate concentration to XYZ matrix
M = alg.dot(T, list(zip(rho[0], rho[1], rho[2])))

# Let's assume full rank. If it isn't, we'll get a singular matrix error.
# If it fails it would require a more advanced approach to try and invert.
pM = alg.pinv(M)  # noqa: N816

# Print out some conversions back and forth so we can see if we get negative results
# above and beyond floating point error.
print('=== Test Conversions ===')
for test in ('#191248', 'red', 'green', 'blue', 'cyan', 'magenta', 'yellow'):
    color = Color(test)
    ans = alg.dot(pM, color.convert('xyz-d65').coords())
    print('Uncorrected', ans)
    ans = [alg.clamp(c, 0, None) for c in ans]
    print('Corrected', ans)
    print(color.to_string(), color[:-1])
    color2 = Color('xyz-d65', alg.dot(M, ans)).convert(color.space(), in_place=True)
    print(color2.to_string(), color2[:-1])

print('=== Concentrations to XYZ ===')
alg.pprint(M)

print('=== XYZ to Concentrations ===')
alg.pprint(pM)

print('=== CMFs D65 ===')
alg.pprint(T)

# Setup plot for results
plt.style.use('seaborn-v0_8-darkgrid')
# Create axes
ax = plt.axes(
    xlabel='Wavelength',
    ylabel='Reflection'
)

for target, r in zip(targets, rho):
    print(f'==== Curve for {target} ====')
    alg.pprint(r)
    plt.plot(
        list(range(START, END, STEP)),
        r,
        color=target,
        marker="",
        linewidth=1.5,
        markersize=2,
    antialiased=True
)

plt.gcf().set_dpi(200)
plt.show()



"""Test UVW 1964."""
import unittest
from . import util
from coloraide_extras.everything import ColorAll as Color
import pytest


class TestUVW(util.ColorAssertsPyTest):
    """Test UVW 1964."""

    COLORS = [
        ('red', 'color(--uvw 171.8 24.715 52.261)'),
        ('orange', 'color(--uvw 73.871 48.705 73.965)'),
        ('yellow', 'color(--uvw 7.628 70.501 96.177)'),
        ('green', 'color(--uvw -42.842 36.934 45.249)'),
        ('blue', 'color(--uvw -9.1161 -84.255 31.317)'),
        ('indigo', 'color(--uvw 9.5941 -38.921 19.483)'),
        ('violet', 'color(--uvw 51.118 -44.069 68.724)'),
        ('white', 'color(--uvw 0 0 99.04)'),
        ('gray', 'color(--uvw 0 0 52.609)'),
        ('black', 'color(--uvw 0 0 -17)'),
        # Test color
        ('color(--uvw 50 10 -10)', 'color(--uvw 50 10 -10)'),
        ('color(--uvw 50 10 -10 / 0.5)', 'color(--uvw 50 10 -10 / 0.5)'),
        ('color(--uvw 50% 50% -50% / 50%)', 'color(--uvw 50 50 -50 / 0.5)'),
        ('color(--uvw none none none / none)', 'color(--uvw none none none / none)'),
        # Test percent ranges
        ('color(--uvw 0% 0% 0%)', 'color(--uvw 0 0 0)'),
        ('color(--uvw 100% 100% 100%)', 'color(--uvw 100 100 100)'),
        ('color(--uvw -100% -100% -100%)', 'color(--uvw -100 -100 -100)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_uvw_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('uvw'), Color(color2))


class TestUVWSerialize(util.ColorAssertsPyTest):
    """Test UVW serialization."""

    COLORS = [
        # Test color
        ('color(--uvw 75 10 -10 / 0.5)', {}, 'color(--uvw 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--uvw 75 10 -10)', {'alpha': True}, 'color(--uvw 75 10 -10 / 1)'),
        ('color(--uvw 75 10 -10 / 0.5)', {'alpha': False}, 'color(--uvw 75 10 -10)'),
        # Test None
        ('color(--uvw none 10 -10)', {}, 'color(--uvw 0 10 -10)'),
        ('color(--uvw none 10 -10)', {'none': True}, 'color(--uvw none 10 -10)'),
        # Test Fit
        ('color(--uvw 75 102 -10)', {}, 'color(--uvw 75 102 -10)'),
        ('color(--uvw 75 102 -10)', {'fit': False}, 'color(--uvw 75 102 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestUVWPoperties(util.ColorAsserts, unittest.TestCase):
    """Test UVW."""

    def test_u(self):
        """Test `u`."""

        c = Color('color(--uvw 7.628 70.501 96.177)')
        self.assertEqual(c['u'], 7.628)
        c['u'] = 0.2
        self.assertEqual(c['u'], 0.2)

    def test_v(self):
        """Test `v`."""

        c = Color('color(--uvw 7.628 70.501 96.177)')
        self.assertEqual(c['v'], 70.501)
        c['v'] = 0.1
        self.assertEqual(c['v'], 0.1)

    def test_w(self):
        """Test `w`."""

        c = Color('color(--uvw 7.628 70.501 96.177)')
        self.assertEqual(c['w'], 96.177)
        c['w'] = 0.1
        self.assertEqual(c['w'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--uvw 7.628 70.501 96.177)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Color test mix-in."""


class ColorAsserts:
    """Asserts."""

    def assertCompare(self, value1, value2, rounding=5):
        """Compare with rounding."""

        self.assertEqual(round(value1, rounding), round(value2, rounding))

    def assertColorEqual(self, color1, color2, fit=False, precision=5):
        """Compare two colors."""

        self.assertEqual(
            color1.to_string(fit=fit, precision=precision),
            color2.to_string(fit=fit, precision=precision)
        )

    def assertColorNotEqual(self, color1, color2, fit=False, precision=5):
        """Compare two colors."""

        self.assertNotEqual(
            color1.to_string(fit=fit, precision=precision),
            color2.to_string(fit=fit, precision=precision)
        )


class ColorAssertsPyTest(ColorAsserts):
    """Asserts for `pytest`."""

    def assertEqual(self, a, b):
        """Assert equal."""

        assert a == b, "{} != {}".format(a, b)

    def assertNotEqual(self, a, b):
        """Assert not equal."""

        assert a != b, "{} != {}".format(a, b)



"""
CIE 1964 UVW class.

https://en.wikipedia.org/wiki/CIE_1964_color_space
"""
from __future__ import annotations
from coloraide.spaces import Space
from coloraide.channels import Channel, FLG_MIRROR_PERCENT
from coloraide.cat import WHITES
from coloraide import util
from coloraide import algebra as alg
from coloraide.types import Vector, VectorLike


def xyz_to_uvw(xyz: Vector, white: VectorLike) -> Vector:
    """
    Translate XYZ to UVW.

    When translating XYZ to the intermediate xyY, we need to treat Y as Y * 100.
    """

    u, v = util.xy_to_uv_1960(util.xyz_to_xyY(xyz, white)[:2])
    u0, v0 = util.xy_to_uv_1960(white)
    w = 25.0 * alg.nth_root(xyz[1] * 100.0, 3) - 17.0
    return [
        13 * w * (u - u0),
        13 * w * (v - v0),
        w
    ]


def uvw_to_xyz(uvw: Vector, white: VectorLike) -> Vector:
    """
    Translate UVW to XYZ.

    When translating xyY back to XYZ, we need to scale Y back as well: Y / 100.
    """

    u0, v0 = util.xy_to_uv_1960(white)
    w = uvw[2]
    x, y = util.uv_1960_to_xy(
        [
            (uvw[0] / (13 * w)) + u0 if w != 0 else u0,
            (uvw[1] / (13 * w)) + v0 if w != 0 else v0
        ]
    )
    return util.xy_to_xyz([x, y], (((w + 17.0) / 25.0) ** 3) / 100.0)


class UVW(Space):
    """The UVW class."""

    BASE = "xyz-d65"
    NAME = "uvw"
    SERIALIZE = ("--uvw",)  # type: tuple[str, ...]
    CHANNELS = (
        Channel("u", -100.0, 100.0, flags=FLG_MIRROR_PERCENT),
        Channel("v", -100.0, 100.0, flags=FLG_MIRROR_PERCENT),
        Channel("w", 0.0, 100.0, flags=FLG_MIRROR_PERCENT)
    )
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return uvw_to_xyz(coords, self.white())

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return xyz_to_uvw(coords, self.white())



"""Test contrast."""
import unittest
from coloraide_extras.everything import ColorAll as Color
from . import util


class TestContrastWeber(util.ColorAsserts, unittest.TestCase):
    """Test Weber contrast ratio specifics."""

    def test_contrast_same(self):
        """Test contrast of to same colors."""

        self.assertEqual(Color('blue').contrast('blue', method='weber'), 0)

    def test_contrast_zero_luminance(self):
        """Test contrast of two blacks (divide by zero)."""

        self.assertEqual(Color('black').contrast('black', method='weber'), 50000)

    def test_contrast_bigger(self):
        """Test greater contrast."""

        self.assertCompare(Color('white').contrast('blue', method='weber'), 12.85189)

    def test_symmetry(self):
        """Test symmetry."""

        self.assertEqual(
            Color('white').contrast('blue', method='weber'),
            Color('blue').contrast('white', method='weber')
        )


class TestContrastMichelson(util.ColorAsserts, unittest.TestCase):
    """Test Michelson contrast ratio specifics."""

    def test_contrast_same(self):
        """Test contrast of two same colors."""

        self.assertEqual(Color('blue').contrast('blue', method='michelson'), 0)

    def test_contrast_zero(self):
        """Test contrast of two black colors (divide by zero)."""

        self.assertEqual(Color('black').contrast('black', method='michelson'), 0)

    def test_contrast_bigger(self):
        """Test greater contrast."""

        self.assertCompare(Color('white').contrast('blue', method='michelson'), 0.86534)

    def test_symmetry(self):
        """Test symmetry."""

        self.assertEqual(
            Color('white').contrast('blue', method='michelson'),
            Color('blue').contrast('white', method='michelson')
        )
